import numpy as np
import theano.tensor as T


def log_stability(x):
    """Log-stability for computing loss
    
    :param x: Input value
    :type x: float
    :return: Scaled value where :math:`\hat{x} \in (0, 1)`
    :rtype: float
    """
    if x==0: 
        return 10e-9 
    elif x==1: 
        return 1.0-10e-9
    else:
        return x


def binary_cross_entropy(y_true, y_pred, tensor=False):
    """Binary cross-entropy
    
    :param y_true: Real values on which to compare.
    :type y_true: numpy.array
    :param y_pred: Predicted values.
    :type y_pred: numpy.array
    :param tensor: Perform computation as tensor (theano type), defaults to False
    :type tensor: bool, optional
    :formula: :math:`loss = y_{i} \log \left[ \hat{y}_{i} \\right] + (1 - y_{i}) \log \left[1 - \hat{y}_{i} \\right]`
    :references: * Friedman, J., Hastie, T. and Tibshirani, R., 2001. `The elements of statistical learning <https://web.stanford.edu/~hastie/Papers/ESLII.pdf>`_. Ch. 2, pp. 24.
    :return: Binary cross-entropy
    :rtype: float
    """
    if tensor:
        loss = -y_true * T.log(y_pred) - (1-y_true) * T.log(1-y_pred)
    else:
        loss = (-y_true * np.log(y_pred) - (1-y_true) * np.log(1-y_pred)).sum()
    # print(loss)
    return loss




def mean_squared_error(y_true, y_pred, root=False):
    """Mean Squared Error
    
    :param y_true: Real values on which to compare.
    :type y_true: numpy.array
    :param y_pred: Predicted values.
    :type y_pred: numpy.array
    :param root: Return Root Mean Squared Error (RMSE), defaults to False.
    :type root: bool, optional

    :formula: :math:`loss = \\dfrac{1}{m} \\times \\sum_{i=1}^{m} (y_i - \\hat{y}_i)^2`

    :references: * Friedman, J., Hastie, T. and Tibshirani, R., 2001. `The elements of statistical learning <https://web.stanford.edu/~hastie/Papers/ESLII.pdf>`_. Ch. 2, pp. 24.

    :return: Mean Squared Error or its root.
    :rtype: float
    """
    loss = ((y_pred - y_true).sum())**2
    if root:
        return loss ** (1/2)
    else:
        return loss



def binary_accuracy(y_true, y_pred):
    """Accuracy for binary data
    
    :param y_true: Real values on which to compare.
    :type y_true: numpy.array
    :param y_pred: Predicted values
    :type y_pred: numpy.array
    
    :return: Binary accuracy (in percent)
    :rtype: float
    """
    return (y_true == y_pred).mean()


# MAPE formula
def mape(y_true, y_pred):
    """Computes the Mean Absolute Percentage Error
    
    :param y_true: Real values on which to compare.
    :type y_true: numpy.array
    :param y_pred: Predicted values.
    :type y_pred: numpy.array
    :formula: :math:`\dfrac{100}{n} \sum_{i=1}^{n} \dfrac{|y - \hat{y}|}{y}`
    
    :return: Mean Absolute Percentage Error as percentage.
    :rtype: float
    """
    y = np.array(y_true)
    yhat = np.array(y_pred)
    m = len(y)
    mape = (100/m) * sum(np.abs(y-yhat))/sum(y)
    return mape



import numpy as np
import pandas as pd

def generate_dataset(coeffs, n, std_dev, intercept=0., distribution='normal', binary=False, seed=None, **kwargs):
    """Generate an artificial dataset

    :param coeffs: List of coefficients to use for computing the ouytput variable.
    :type coeffs: :obj:`list`
    :param n: Number of observations to generate.
    :type n: :obj:`int`
    :param std_dev: Standard deviation of the distribution.
    :type std_dev: :obj:`list`
    :param intercept: Value of the intercept to be set, defaults to 0.
    :type intercept: :obj:`float`, optional
    :param distribution: Type of distribution to use for generating the input variables, defaults to 'normal'. Can be:

        * `normal`: :math:`X \\sim \\mathcal{N}(\\mu, \\sigma^{2})`
        * `unirform`: :math:`X \\sim \\mathcal{U}_{[\\text{low}, \\text{high}]}`

    :type distribution: :obj:`str`, optional
    :param binary: Define if output is binary, defaults to False.
    :type binary: :obj:`bool`, optional
    :param seed: Random seed, defaults to None.
    :type seed: :obj:`int`, optional

    :param \*\*kwargs: Arguments to be passed in the distribution function. Can be:

        * `normal`: :obj:`loc` = :math:`\\mu` and :obj:`scale` = :math:`\\sigma^{2}`
        * `uniform`: :obj:`low` and :obj:`high`

    :return: DataFrame with output variable named as :obj:`Y` and covariates as :obj:`X0`, :obj:`X1`, :obj:`X2`, ...
    :rtype: :obj:`pandas.DataFrame`
    """

    rdm = np.random.RandomState(seed) if seed else np.random
    # We calculate the number of predictors, and create a coefficient matrix
    # With `p` rows and 1 column, for matrix multiplication
    p = len(coeffs)
    params = pd.DataFrame({'coeff': coeffs, 'std_dev': std_dev})

    # Similar as before, but with `n` rows and `p` columns this time
    x = []
    for index, row in params.iterrows():
        if distribution.lower() == 'normal':
            x += [rdm.normal(size=n, **kwargs)]
        if distribution.lower() == 'uniform':
            x += [rdm.uniform(size=n, **kwargs)]
    X = np.array(x)
    e = rdm.normal(loc=0., scale=1., size=n)

    # Since x is a n*p matrix, and coefficients is a p*1 matrix
    # we can use matrix multiplication to get the value of y for each
    # set of values x1, x2 .. xp
    # We need to transpose it to get a 1*n array from a n*1 matrix to use in the regression model
    y = X.T.dot(coeffs) + e + intercept

    if binary:
        y = [1 if y_i > 0 else 0 for y_i in y]
    
    df = pd.DataFrame(X.T)
    # Assign column names
    for col in df.columns:
        df.rename(columns = {col: 'X' + str(col)}, inplace=True)
    
    # Append Y
    df['Y'] = y
    #
    return df



import numpy as np
import random
import theano
import theano.tensor as T
import matplotlib.pyplot as plt

# import pycof as pc

import sys
sys.path.append("/Users/flofelic/Documents/statinf/")

from statinf.ml.losses import binary_cross_entropy
from statinf.ml.activations import logit
from statinf.data.GenerateData import generate_dataset



class GLM:
    def __init__(self, formula, data, fit_intercept=False, test_set=None, training_size=0.2, seed=None):
        super(GLM, self).__init__()
        # Parse formula
        self.no_space_formula = formula.replace(' ', '')
        self.Y_col = self.no_space_formula.split('~')[0]
        self.X_col = self.no_space_formula.split('~')[1].split('+')
        # Data information
        self.data = data
        self.N = data.shape[0]
        self.p = len(self.X_col)
        self.test_set = test_set
        # Set randomness if asked
        if seed is not None:
            self.rdm = np.random.RandomState(seed)
            random.seed(seed)
        else:
            self.rdm = np.random

        # Set test data
        if test_set is None:
            # Getting the data indexes
            data_idx = data.index.tolist()
            # Get random subsample from index for training and test sets
            train_index = random.sample(data_idx, int(self.N * training_size))
            test_index = [i for i in data_idx if i not in train_index]
            # Training set
            self.Y_train = [float(y) for y in self.data.loc[train_index, self.Y_col].values]
            self.X_train = self.data.loc[train_index, self.X_col]
            # Test set with randomly selected indexes
            self.Y_test = [float(y) for y in self.data.loc[test_index, self.Y_col].values]
            self.X_test = self.data.loc[test_index, self.X_col]
        else:
            # Training set from initial data provided
            self.Y_train = [float(y) for y in self.test_set[self.Y_col].values]
            self.X_train = self.data[self.X_col]
            # Test set from provided data
            self.Y_test = [float(y) for y in self.test_set[self.Y_col].values]
            self.X_test = self.test_set[self.X_col]
        
        # Theano data format
        ## Train set
        self.Y_train_theano = np.array(self.Y_train).astype(theano.config.floatX)
        self.X_train_theano = self.X_train.values.astype(theano.config.floatX)
        ## Test set
        self.Y_test_theano = np.array(self.Y_test).astype(theano.config.floatX)
        self.X_test_theano = self.X_test.values.astype(theano.config.floatX)

    
    def fit(self, epochs=1000, patience=0.1, improvement_threshold=0.995, learning_rate=0.01, restore_weights=True, early_stop=True, plot=False, verbose=False, verbose_all=False):
        
        verbose = True if verbose_all else verbose
        # Parse arguments
        self.learning_rate = learning_rate
        if type(patience) == float:
            # User can provide either test data size or ratio of total data
            patience = epochs * patience
        
        # Initialize values
        train_losses = []       # Training loss vector to plot
        test_losses = []        # Test loss vector to plot
        self.best_loss = np.inf # Initial best loss value
        verif_i = epochs        # First, verify epoch at then end (init only)
        i = 0

        # Declare Theano symbolic variables
        self.x = T.matrix("x")
        self.y = T.vector("y")
        self.w = theano.shared(self.rdm.randn(self.p).astype(theano.config.floatX), name="w")
        self.b = theano.shared(np.asarray(0., dtype=theano.config.floatX), name="b")
        # Test tag values from Theano objects
        self.x.tag.test_value = self.X_train_theano
        self.y.tag.test_value = self.Y_train_theano
        
        # Construct Theano expression graph
        self.p_1 = logit(x=self.x, weights=self.w, bias=self.b, tensor=True)    # Probability of having a one
        self.prediction = self.p_1 > 0.5                              # The prediction that is done: 0 or 1

        # Compute gradients
        self.xent = binary_cross_entropy(self.y, self.p_1, tensor=True)
        self.cost = self.xent.mean() #+ 0.01*(w**2).sum() # The cost to optimize
        self.gw, self.gb = T.grad(self.cost, [self.w, self.b])

        # Training and prediction functions
        self._train = theano.function(inputs=[self.x, self.y], outputs=[self.prediction, self.xent], updates=[[self.w, self.w-self.learning_rate*self.gw], [self.b, self.b-self.learning_rate*self.gb]], name = "train")
        self._predict = theano.function(inputs=[self.x], outputs=self.prediction, name="predict")

        # Start training
        while i <= epochs:
            i += 1
            # Train
            output = self._train(self.X_train_theano, self.Y_train_theano)
            # Evaluate on test set
            test_out = self._predict(self.X_test_theano)
            # Measure accuracy on train and test sets
            epoch_loss = binary_cross_entropy(self.Y_train_theano, output[0])
            epoch_loss_test = binary_cross_entropy(self.Y_test_theano, test_out)
            # Save to historical performance
            train_losses += [epoch_loss]
            test_losses += [epoch_loss_test]
            # If we got the best score until now, and patience is not reached
            if (epoch_loss_test < self.best_loss * improvement_threshold) & (i <= verif_i) & early_stop:
                # pc.verbose_display(f'New min found for iter {i}', verbose=verbose_all)
                verif_i = i + patience
                self.best_loss = epoch_loss_test
                self.best_iter = i
                self.optimal_w = self.w.get_value()
                self.optimal_b = self.b.get_value()
            # if no improvement
            elif (i <= verif_i):
                # pc.verbose_display(f'No improvement at iter {i}', verbose=verbose_all)
                pass
            else:
                # pc.verbose_display(f'Stop training. Minimum found after {self.best_iter} iterations', verbose)
                break

        if restore_weights:
            # Keep the parameters of the best run and not of the last one
            self.w = theano.shared(self.optimal_w.astype(theano.config.floatX), name="w")
            self.b = theano.shared(self.optimal_b.astype(theano.config.floatX), name="b")
        
        if plot:
            # Plot the training and test loss
            plt.title('Logit train vs. test error through epochs', loc='center')
            plt.plot(list(range(i)), train_losses, label='Training loss')
            plt.plot(list(range(i)), test_losses, label='Test loss')
            plt.legend()
            plt.show()

    def predict(self, new_data):
        data_to_pred = new_data[self.X_col].values.astype(theano.config.floatX)
        new_out = self._predict(data_to_pred)
        return [float(y) for y in new_out]

    def get_weights(self):
        self.parameters = {'weights': self.w.get_value(), 'bias': self.b.get_value()}
        return self.parameters






import numpy as np

from ..ml.initializations import init_params
from ..ml.activations import sigmoid
# from .glm_test import GLM


class GLM:
    """ 
    [C. Cameron & P.k. Trivedi] - Microeconometrics methods and applications - 2005
    """
    def __init__(self, formula, data, fit_intercept=False, initializer='zeros'):
        self.data             = data # Pandas DataFrame
        # Parse the formula
        self.no_space_formula = formula.replace(' ', '')
        self.fit_intercept    = fit_intercept
        self.Y_col            = self.no_space_formula.split('~')[0]
        self.X_col            = self.no_space_formula.split('~')[1].split('+')
        self.X                = data[self.X_col].to_numpy()
        self.Y                = data[self.Y_col].to_numpy()
        self.log_likelihood   = []
        self.gradients        = []
        self.coefs_hist       = []
        self.hessian_hist     = []
        self.jacobian_hist    = []

        # Need to fit intercept?
        if self.fit_intercept:
            self.X.insert(1, 'intercept', 1)
        # Initialize parameters
        self.beta = init_params(rows=len(self.X_col), cols=1, method=initializer, isTheano=False)


    def _explain(self, beta):
        """
        1rst step build the linear combination
        """
        Xb   = self.X.dot(np.array(beta).T)
        return Xb

    def _prob(self, X):
        """ 
        Prob
        """
        prob = sigmoid(X.dot(self.beta))
        # prob = np.array([(1/(1+np.exp(-x))) for x in self._explain(beta)])
        return prob
        
    def _log_likelihood(self, X=None, Y=None):
        """
        Compute Log-Likelihood 
        """
        X_data = self.X if X is None else X
        Y_data = self.Y if Y is None else Y

        # Individual log-likelihood
        log_likelihood_id = np.array(Y_data * np.log(self._prob(X_data)) + (1 - Y_data) * np.log(1-self._prob(X_data)))
        # Overall log-likelihood
        log_likelihood = log_likelihood_id.sum()
        return log_likelihood


    def _jacobian(self):
        """
        Gradient (Overall) not the score function 
        """
        gradient = np.array(self.Y - self._prob(self.X)).dot(self.X) # gradient of the likelihood
        return gradient

    def _hessian(self):
        """
        Hessian -- Fisher Information 
        """
        hessian = np.array((self._prob(self.X).dot(1 - self._prob(self.X)).sum()) * self.X.T.dot(self.X)) #! Wrong formula
        return hessian

    def variance(self):
        """
        Fisher Information
        """
        variance = np.linalg.inv(self._hessian())
        return np.array(variance)

    def fit(self, max_it=10, improvement_threshold=0.995, keep_hist=True):
        it = 0
        diff = np.inf
        old_log_like = np.inf
    
        while (it < max_it) and (np.abs(diff) > 1-improvement_threshold):
            it           += 1
            log_like      = self._log_likelihood()
            gradient      = self._jacobian()
            variation     = self.variance().dot(gradient)
            self.beta    -= variation.T
            diff          = old_log_like - log_like
            old_log_like  = log_like

            if keep_hist:
                self.log_likelihood += [log_like]
                self.gradients      += [gradient]
            
        return print(it)

    def predict(self, max_it=10):
        beta         = self._fit(max_it)
        self.predict = [1 if x > 0.5 else 0 for x in self._prob(beta)]
        return print(f"""{self.predict}""")
    



import theano
import theano.tensor as T
import numpy as np
from collections import OrderedDict


"""
References
----------
https://github.com/satopirka/deep-learning-theano/blob/master/example/mlp.py
https://github.com/satopirka/deep-learning-theano/blob/master/dnn/optimizers.py
"""

def build_shared_zeros(shape, name='v'):
    """Builds a theano shared variable filled with a zeros numpy array
    
    :param shape: Shape of the vector to create.
    :type shape: tuple
    :param name: Name to give to the shared theano value, defaults to 'v'.
    :type name: str

    :example:

    >>> build_shared_zeros(shape=(5, 10), name='r')

    :return: Theano shared matrix
    :rtype: theano.shared
    """
    return theano.shared(
        value=np.zeros(shape, dtype=theano.config.floatX), 
        name=name, 
        borrow=True
    )


class Optimizer(object):
    """Optimization updater
    
    :param object: Optimizer object
    :type object: class
    """
    def __init__(self, params=None):
        if params is None:
            raise NotImplementedError()
        self.params = params

    def updates(self, loss=None):
        if loss is None:
            raise NotImplementedError()

        self.updates = OrderedDict()
        self.gparams = [T.grad(loss, param) for param in self.params]


class SGD(Optimizer):
    """Stochastic Gradient Descent optimizer
    
    :param params: List compiling the parameters to update.
    :type params: list
    :param learning_rate: Step size, defaults to 0.01.
    :type learning_rate: float

    :formula: .. math:: \\theta_{t} = \\theta_{t-1} - \\epsilon \\nabla f_{t}(\\theta_{t-1})

    :references: Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:
    
    >>> SGD(params=[W, b], learning_rate=0.01).updates(cost)
    """
    def __init__(self, learning_rate=0.01, params=None):
        super(SGD, self).__init__(params=params)
        self.learning_rate = learning_rate

    def updates(self, loss=None):
        """Update loss using the SGD formula
        
        :param loss: Vector of loss to be updated, defaults to None
        :type loss: tensor, optional

        :return: Updated parameters and gradients
        :rtype: tensor
        """
        super(SGD, self).updates(loss=loss)

        for param, gparam in zip(self.params, self.gparams):
            self.updates[param] = param - self.learning_rate * gparam

        return self.updates


class MomentumSGD(Optimizer):
    """Stochastic Gradient Descent with Momentum
    
    :param params: List compiling the parameters to update.
    :type params: list
    :param learning_rate: Step size, defaults to 0.01.
    :type learning_rate: float
    :param alpha: Momentum parameter, defaults to 0.9.
    :type alpha: float

    :formula: .. math:: \\theta_{t} = \\theta_{t-1} + \\alpha v - \\epsilon \\nabla f_{t}(\\theta_{t-1})

    :references: Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> MomentumSGD(params=[W, b], learning_rate=0.01, alpha=0.9).updates(cost)
    """
    def __init__(self, params=None, learning_rate=0.01, alpha=0.9):
        super(MomentumSGD, self).__init__(params=params)
        self.learning_rate = learning_rate
        # Momentum value
        self.alpha = alpha
        self.vs = [build_shared_zeros(t.shape.eval(), 'v') for t in self.params]

    def updates(self, loss=None):
        """Update loss using the Momentum SGD formula
        
        :param loss: Vector of loss to be updated, defaults to None
        :type loss: tensor, optional

        :return: Updated parameters and gradients
        :rtype: tensor
        """
        super(MomentumSGD, self).updates(loss=loss)

        for v, param, gparam in zip(self.vs, self.params, self.gparams):
            _v = self.alpha * v - self.learning_rate * gparam
            self.updates[param] = param + _v
            self.updates[v] = _v

        return self.updates


class RMSprop(Optimizer):
    """RMSprop optimizer
    
    :param params: List compiling the parameters to update.
    :type params: list
    :param learning_rate: Step size, defaults to 0.001.
    :type learning_rate: float
    :param rho: Decay rate, defaults to 0.9.
    :type rho: float
    :param delta: Constant for division stability, defaults to 10e-6.
    :type delta: float

    :formula: .. math:: r = \\rho r + (1- \\rho) \\nabla f_{t}(\\theta_{t-1}) \\odot \\nabla f_{t}(\\theta_{t-1})
    
        .. math:: \\theta_{t} = \\theta_{t-1} - \\dfrac{\\epsilon}{\\sqrt{\\delta + r}} \\odot \\nabla f_{t}(\\theta_{t-1})


    :references: * Tieleman, T., & Hinton, G. (2012). `Lecture 6.5-rmsprop: Divide the gradient by a running average of its recent magnitude <https://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf>`_. COURSERA: Neural networks for machine learning, 4(2), 26-31.
        
        * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> RMSprop(params=[W, b], learning_rate=0.01, rho=0.9, delta=10e-6).updates(cost)
    """
    def __init__(self, params=None, learning_rate=0.001, rho=0.9, delta=10e-6):
        super(RMSprop, self).__init__(params=params)

        self.learning_rate = learning_rate
        self.rho = rho
        self.delta = delta

        self.mss = [build_shared_zeros(t.shape.eval(), 'r') for t in self.params]

    def updates(self, loss=None):
        """Update loss using the RMSprop formula
        
        :param loss: Vector of loss to be updated, defaults to None
        :type loss: tensor, optional

        :return: Updated parameters and gradients
        :rtype: tensor
        """
        super(RMSprop, self).updates(loss=loss)

        for r, param, gparam in zip(self.mss, self.params, self.gparams):
            _r = r * self.rho
            _r += (1 - self.rho) * gparam * gparam
            self.updates[r] = _r
            self.updates[param] = param - self.learning_rate * gparam / T.sqrt(self.delta + _r)

        return self.updates


class Adam(Optimizer):
    """Adaptive Moments optimizer
    
    :param params: List compiling the parameters to update.
    :type params: list
    :param learning_rate: Step size, defaults to 0.001.
    :type learning_rate: float
    :param beta1: Exponential decay rate for first moment estimate, defaults to 0.9.
    :type beta1: float
    :param beta2: Exponential decay rate for scond moment estimate, defaults to 0.999.
    :type beta2: float
    :param delta: Constant for division stability, defaults to 10e-8.
    :type delta: float

    :formula: .. math:: m_{t} = \\beta_{1} m_{t-1} + (1 - \\beta_{1}) \\nabla_{\\theta} f_{t}(\\theta_{t-1})
            .. math:: v_{t} = \\beta_{2} v_{t-1} + (1 - \\beta_{2}) \\nabla_{\\theta}^{2} f_{t}(\\theta_{t-1})
            .. math:: \\hat{m}_{t} = \\dfrac{m_{t}}{1 - \\beta_{1}^{t}}
            .. math:: \\hat{v}_{t} = \\dfrac{v_{t}}{1 - \\beta_{2}^{t}}
            .. math:: \\theta_{t} = \\theta_{t-1} - \\alpha \\dfrac{\\hat{m}_{t}}{\\sqrt{\\hat{v}_{t}} + \\delta}


    :references: * Kingma, D. P., & Ba, J. (2014). `Adam: A method for stochastic optimization <https://arxiv.org/pdf/1412.6980.pdf>`_. arXiv preprint arXiv:1412.6980
        
        * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> ADAM(params=[W, b], learning_rate=0.001, beta1=0.9, beta2=0.999, delta=10e-8).updates(cost)
    """
    def __init__(self, params=None, learning_rate=0.001, beta1=0.9, beta2=0.999, delta=10e-8):
        super(Adam, self).__init__(params=params)

        self.alpha = learning_rate
        self.b1 = beta1
        self.b2 = beta2
        self.t = theano.shared(np.float32(1))
        self.delta = delta

        self.ms = [build_shared_zeros(t.shape.eval(), 'm') for t in self.params]
        self.vs = [build_shared_zeros(t.shape.eval(), 'v') for t in self.params]

    def updates(self, loss=None):
        """Update loss using the Adam formula
        
        :param loss: Vector of loss to be updated, defaults to None
        :type loss: tensor, optional

        :return: Updated parameters and gradients
        :rtype: tensor
        """
        super(Adam, self).updates(loss=loss)

        for m, v, param, gparam in zip(self.ms, self.vs, self.params, self.gparams):
            _m = self.b1 * m + (1 - self.b1) * gparam
            _v = self.b2 * v + (1 - self.b2) * gparam ** 2

            m_hat = _m / (1 - self.b1 ** self.t)
            v_hat = _v / (1 - self.b2 ** self.t)

            self.updates[param] = param - self.alpha * m_hat / (T.sqrt(v_hat) + self.delta)
            self.updates[m] = _m
            self.updates[v] = _v
        self.updates[self.t] = self.t + 1.0

        return self.updates


class AdaGrad(Optimizer):
    """Adaptive Gradient optimizer
    
    :param params: List compiling the parameters to update.
    :type params: list
    :param learning_rate: Step size, defaults to 0.01.
    :type learning_rate: float
    :param delta: Constant for division stability, defaults to 1e-7.
    :type delta: float

    :formula: .. math:: r = r + \\nabla f_{t}(\\theta_{t}) \\odot \\nabla f_{t}(\\theta_{t-1})
            .. math:: \\theta_{t} = \\theta_{t-1} - \\frac{\\epsilon}{\\sqrt{\\delta + r}} \\odot \\nabla f_{t}(\\theta_{t-1})


    :references: * Duchi, J., Hazan, E., & Singer, Y. (2011). `Adaptive subgradient methods for online learning and stochastic optimization <http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf>`_. Journal of machine learning research, 12 (Jul), 2121-2159.
        * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> AdaGrad(params=[W, b], learning_rate=0.01, delta=1e-7).updates(cost)
    """
    def __init__(self, params=None, learning_rate=0.01, delta=1e-7):
        super(AdaGrad, self).__init__(params=params)

        self.learning_rate = learning_rate
        # Small constant for numerical stability
        self.delta = delta
        # Gradient accumulation variable
        self.accugrads = [build_shared_zeros(t.shape.eval(), 'r') for t in self.params]

    def updates(self, loss=None):
        """Update loss using the AdaGrad formula
        
        :param loss: Vector of loss to be updated, defaults to None
        :type loss: tensor, optional

        :return: Updated parameters and gradients
        :rtype: tensor
        """
        super(AdaGrad, self).updates(loss=loss)

        for r, param, gparam in zip(self.accugrads, self.params, self.gparams):
            _r = r + gparam * gparam
            self.updates[param] = param - (self.learning_rate / T.sqrt(_r + self.delta)) * gparam
            self.updates[r] = _r

        return self.updates


class AdaMax(Optimizer):
    """AdaMax optimizer (Adam with infinite norm)
    
    :param params: List compiling the parameters to update.
    :type params: list
    :param learning_rate: Step size, defaults to 0.001.
    :type learning_rate: float
    :param beta1: Exponential decay rate for first moment estimate, defaults to 0.9.
    :type beta1: float
    :param beta2: Exponential decay rate for scond moment estimate, defaults to 0.999.
    :type beta2: float

    :formula: .. math:: m_{t} = \\beta_{1} m_{t-1} + (1 - \\beta_{1}) \\nabla_{\\theta} f_{t}(\\theta_{t-1})
            .. math:: u_{t} = \\max(\\beta_{2} \\cdot u_{t-1}, |\\nabla f_{t}(\\theta_{t-1})|)
            .. math:: \\theta_{t} = \\theta_{t-1} - \\alpha \\dfrac{\\hat{m}_{t}}{\\sqrt{\\hat{v}_{t}} + \\epsilon}


    :references: * Kingma, D. P., & Ba, J. (2014). `Adam: A method for stochastic optimization <https://arxiv.org/pdf/1412.6980.pdf>`_. arXiv preprint arXiv:1412.6980
        * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press.

    :example:

    >>> AdaMax(params=[W, b], learning_rate=0.001, beta1=0.9, beta2=0.999).updates(cost)
    """
    
    def __init__(self, params=None, learning_rate=0.001, beta1=0.9, beta2=0.999):
        super(AdaMax, self).__init__(params=params)

        self.alpha = learning_rate
        self.b1 = beta1
        self.b2 = beta2
        self.t = theano.shared(np.float32(1))

        self.ms = [build_shared_zeros(t.shape.eval(), 'm') for t in self.params]
        self.us = [build_shared_zeros(t.shape.eval(), 'u') for t in self.params]

    def updates(self, loss=None):
        """Update loss using the AdaMax formula
        
        :param loss: Vector of loss to be updated, defaults to None
        :type loss: tensor, optional

        :return: Updated parameters and gradients
        :rtype: tensor
        """
        super(AdaMax, self).updates(loss=loss)

        for m, u, param, gparam in zip(self.ms, self.us, self.params, self.gparams):
            _m = self.b1 * m + (1 - self.b1) * gparam
            _u = T.maximum(self.b2 * u, T.abs_(gparam))

            self.updates[param] = param - (self.alpha / (1 - self.b1 ** self.t)) * (_m / _u)
            self.updates[m] = _m
            self.updates[u] = _u
        self.updates[self.t] = self.t + 1.0

        return self.updates



import numpy as np
from scipy import stats
import pandas as pd


#TODO: Add Log-Likehood + AIC + BIC
#TODO: Add dask for GPU usage


class OLS:
    
    def __init__(self, formula, data, fit_intercept=True):
        """Ordinary Least Squares regression
        
        :param formula:  Regression formula to be run of the form :obj:`y ~ x1 + x2`.
        :type formula: :obj:`str`
        :param data: Input data with Pandas format.
        :type data: :obj:`pandas.DataFrame`
        :param fit_intercept: Used for adding intercept in the regression formula, defaults to True.
        :type fit_intercept: :obj:`bool`, optional
        """

        super(OLS, self).__init__()
        # Parse formula
        self.no_space_formula = formula.replace(' ', '')
        self.Y_col = self.no_space_formula.split('~')[0]
        self.X_col = self.no_space_formula.split('~')[1].split('+')
        # Subset X
        self.X = data[self.X_col].to_numpy()
        # Target variable
        self.Y = data[self.Y_col].to_numpy()
        # Degrees of freedom of the population
        self.dft = self.X.shape[0]
        # Degree of freedom of the residuals
        self.dfe = self.X.shape[0] - self.X.shape[1]
        # Size of the population
        self.n = self.X.shape[0]
        # Number of explanatory variables / estimates
        self.p = self.X.shape[1]
        # Use intercept or only explanatory variables
        self.fit_intercept = fit_intercept
    
    def _get_X(self):
        if self.fit_intercept:
            return(np.hstack((np.ones((self.n, 1), dtype=self.X.dtype), self.X)))
        else:
            return(self.X)
    
    def get_betas(self):
        """Computes the estimates for each explanatory variable

        :formula: .. math:: \\beta = (X'X)^{-1} X'Y

        :return: Estimated coefficients
        :rtype: :obj:`numpy.array`
        """
        XtX = self._get_X().T.dot(self._get_X())
        XtX_1 = np.linalg.inv(XtX)
        XtY = self._get_X().T.dot(self.Y)
        beta = XtX_1.dot(XtY)
        return(beta)
    
    def fitted_values(self):
        """Computes the estimated values of Y

        :formula: .. math:: \\hat{Y} = X \\beta

        :return: Fitted values for Y.
        :rtype: :obj:`numpy.array`
        """
        betas = self.get_betas()
        Y_hat = np.zeros(len(self.Y))
        for i in range(len(betas)):
            Y_hat += (betas[i] * self._get_X().T[i])
        return(Y_hat)
    
    def _get_error(self):
        """Compute the error term/residuals

        :formula: .. math:: \\epsilon = Y - \\hat{Y}

        :return: Estimated residual term.
        :rtype: :obj:`numpy.array`
        """
        res = self.Y - self.fitted_values()
        return(res)

    def rss(self):
        """Residual Sum of Squares

        :formula: .. math:: RSS = \\sum_{i=1}^{n} (y_{i} - \\hat{y}_{i})^{2}
            
            where :math:`y_{i}` denotes the true/observed value of :math:`y` for individual :math:`i` and :math:`\\hat{y}_{i}` denotes the predicted value of :math:`y` for individual :math:`i`.

        :return: Residual Sum of Squares.
        :rtype: :obj:`float`
        """
        return((self._get_error()**2).sum())
    
    def tss(self):
        """Total Sum of Squares

        :formula: .. math:: TSS = \\sum_{i=1}^{n} (y_{i} - \\bar{y})^{2}
            
            where :math:`y_{i}` denotes the true/observed value of :math:`y` for individual :math:`i` and :math:`\\bar{y}_{i}` denotes the average value of :math:`y`.

        :return: Total Sum of Squares.
        :rtype: :obj:`float`
        """
        
        y_bar = self.Y.mean()
        total_squared = (self.Y - y_bar) ** 2
        return(total_squared.sum())
    
    def r_squared(self):
        """:math:`R^{2}` -- Goodness of fit

        :formula: .. math:: R^{2} = 1 - \dfrac{RSS}{TSS}

        :return: Goodness of fit.
        :rtype: :obj:`float`
        """

        return(1 - self.rss()/self.tss())

    def adjusted_r_squared(self):
        """Adjusted-:math:`R^{2}` -- Goodness of fit

        :formula: .. math:: R^{2}_{adj} = 1 - (1 - R^{2}) \dfrac{n - 1}{n - p - 1}

            where :math:`p` denotes the number of estimates (i.e. explanatory variables) and :math:`n` the sample size

        :references: Theil, Henri (1961). Economic Forecasts and Policy.

        :return: Adjusted goodness of fit.
        :rtype: :obj:`float`
        """

        adj_r_2 = 1 - (1 - self.r_squared()) * (self.n - 1) / (self.n - self.p - 1)
        return(adj_r_2)
    
    def _fisher(self):
        """Fisher test

        :formula: .. math:: \\mathcal{F} = \dfrac{TSS - RSS}{\\frac{RSS}{n - p}}

            where :math:`p` denotes the number of estimates (i.e. explanatory variables) and :math:`n` the sample size

        :references: Shen, Q., & Faraway, J. (2004). `An F test for linear models with functional responses <https://www.jstor.org/stable/24307230>`_. Statistica Sinica, 1239-1257.

        :return: Value of the :math:`\\mathcal{F}`-statistic.
        :rtype: :obj:`float`
        """
        MSE = (self.tss() - self.rss()) / (self.p - 1)
        MSR = self.rss() / self.dfe
        return(MSE/MSR)
    
    """
        Returns statistics summary for estimates
        
        Formula
        -------
        The p-values are computes as:
        p_value = 2 * (1 - T_n(t_value))
        
         * T denotes the Student Cumulative Distribution Function with n degrees of freedom
        
        The covariance matrix of beta is compute by:
        Var(b) = sigma_b**2 * (X'X)^-1
        
         * sigma_b denotes the standard deviation computed for a given estimate b
                
        Reference
        ---------
        
        """
    def summary(self, return_df=False):
        """Statistical summary for OLS
        
        :param return_df: Return the summary as a Pandas DataFrame, else print a string, defaults to False.
        :type return_df: :obj:`bool`

        :formulae: * Fisher test:

            .. math:: \\mathcal{F} = \dfrac{TSS - RSS}{\\frac{RSS}{n - p}}

            where :math:`p` denotes the number of estimates (i.e. explanatory variables) and :math:`n` the sample size


            * Covariance matrix:

            .. math:: \\mathbb{V}(\\beta) = \\sigma^{2} X'X

            where :math:`\\sigma^{2} = \\frac{RSS}{n - p -1}`


            * Coefficients' significance:

            .. math:: p = 2 \\left( 1 - T_{n} \\left( \\dfrac{\\beta}{\\sqrt{\\mathbb{V}(\\beta)}} \\right) \\right)

            where :math:`T` denotes the Student cumulative distribution function (c.d.f.) with :math:`n` degrees of freedom


        :references: * Student. (1908). The probable error of a mean. Biometrika, 1-25.
            * Shen, Q., & Faraway, J. (2004). `An F test for linear models with functional responses <https://www.jstor.org/stable/24307230>`_. Statistica Sinica, 1239-1257.
            * Wooldridge, J. M. (2016). `Introductory econometrics: A modern approach <https://faculty.arts.ubc.ca/nfortin/econ495/IntroductoryEconometrics_AModernApproach_FourthEdition_Jeffrey_Wooldridge.pdf>`_. Nelson Education.
        """
        # Initialize
        betas = self.get_betas()
        X = self._get_X()

        sigma_2 = (sum((self._get_error())**2))/(len(X) - len(X[0]))
        variance_beta = sigma_2 * (np.linalg.inv(np.dot(X.T, X)).diagonal())
        std_err_beta = np.sqrt(variance_beta)
        t_values = betas / std_err_beta
        
        p_values = [2 * (1 - stats.t.cdf(np.abs(i), (len(X)-1))) for i in t_values]
        
        summary_df = pd.DataFrame()
        summary_df["Variables"] = ['(Intercept)'] + self.X_col if self.fit_intercept else self.X_col
        summary_df["Coefficients"] = betas
        summary_df["Standard Errors"] = std_err_beta
        summary_df["t-values"] = t_values
        summary_df["Probabilities"] = p_values
        
        r2 = self.r_squared()
        adj_r2 = self.adjusted_r_squared()
        #
        fisher = self._fisher()
        #
        if return_df:
            return(summary_df)
        else:
            print('=========================================================================')
            print('                               OLS summary                               ')
            print('=========================================================================')
            print('| R² = {:.5f}                  | Adjusted-R² = {:.5f}'.format(r2, adj_r2))
            print('| n  = {:6}                   | p = {:5}'.format(self.n, self.p))
            print('| Fisher = {:.5f}                         '.format(fisher))
            print('=========================================================================')
            print(summary_df.to_string(index=False))

    def predict(self, new_data):
        """Predicted :math:`\\hat{Y}` values for for a new dataset
        
        :param new_data: New data to evaluate with pandas data-frame format.
        :type new_data: :obj:`pandas.DataFrame`
        
        :formula: .. math:: \\hat{Y} = X \\hat{\\beta}

        :return: Predictions
        :rtype: :obj:`numpy.array`
        """
        X_array = new_data[self.X_col].to_numpy()
        if self.fit_intercept:
            new_X = np.hstack((np.ones((new_data.shape[0], 1), dtype=X_array.dtype), X_array))
        else:
            new_X = X_array
        
        return np.dot(new_X, self.get_betas())



import numpy as np
import theano


def init_params(rows, cols, method='xavier', mean=0., std=1., tensor=True, seed=None):
    """Initialize the weight and bias matrices based on probabilistic distribution.
    
    :param rows: Size of the input, number of rows to be generated.
    :type rows: int
    :param cols: Size of the output, number of columns to be generated.
    :type cols: int
    :param method: Distibution to use to generated the weights, defaults to 'xavier'.
    :type method: str, optional
    :param mean: Mean for the distribution to be generated, defaults to 0.
    :type mean: float, optional
    :param std: Standard deviation for the distribution to be generated, defaults to 1.
    :type std: float, optional
    :param tensor: Needs to return a theano friendly-format, defaults to True.
    :type tensor: bool, optional
    :param seed: Seed to be set for randomness, defaults to None.
    :type seed: int, optional
    
    :raises ValueError: `method` needs to be 'ones', 'zeros', 'uniform', 'xavier' or 'normal', see below for details.
    
    :method: * **Zeros**: :math:`W_j = \\vec{0}`
        * **Ones**: :math:`W_j = \\vec{1}`
        * **Uniform**: :math:`W_j \\sim \\mathcal{U} _{\\left[0, 1 \\right)}`
        * **Xavier**: :math:`W_j \\sim \\mathcal{U}\\left[ -\\frac{\\sqrt{6}}{\\sqrt{n_j + n_{j+1}}}, \\frac{\\sqrt{6}}{\\sqrt{n_j + n_{j+1}}} \\right]`
        * **Normal**: :math:`W_j \\sim \\mathcal{N}(0, 1)`

    :references: * Neuneier, Ralph, and Hans Georg Zimmermann. "`How to train neural networks <https://link.springer.com/chapter/10.1007/3-540-49430-8_18>`_" In Neural networks: tricks of the trade, pp. 373-423. Springer, Berlin, Heidelberg, 1998.
        * Glorot, Xavier, and Yoshua Bengio. "`Understanding the difficulty of training deep feedforward neural networks <http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf>`_" In Proceedings of the thirteenth international conference on artificial intelligence and statistics, pp. 249-256. 2010.
    :return: Weight or bias matrix for initiating ML optimization
    :rtype: numpy.array
    """

    # Set random state
    rdm = np.random.RandomState(seed) if seed is not None else np.random

    # Get the weights
    if method.upper() == 'ONES':
        W = np.ones((rows, cols))
    if method.upper() == 'ZEROS':
        W = np.zeros((rows, cols))
    elif method.upper() == 'UNIFORM':
        W = rdm.rand(rows, cols)
    elif method.upper() == 'XAVIER':
        W = rdm.uniform(low = -np.sqrt(6. / (rows + cols)),
                    high = np.sqrt(6. / (rows + cols)),
                    size = (rows, cols))
    elif method.upper() == 'NORMAL':
        W = rdm.normal(mean, std, rows * cols).reshape((rows, cols))
    else:
        raise ValueError('Weight initialization method not valid.')

    if tensor:
        return np.asarray(W, dtype=theano.config.floatX)
    else:
        return W




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="statinf",
    version="1.0.1",
    author="Florian Felice",
    author_email="florian.felice@outlook.com",
    description="A library for statistics and causal inference",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/florianfelice/statinf",
    packages=setuptools.find_packages(),
    install_requires=[
          "pandas>=0.24.1",
          "numpy>=1.16.3",
          "scipy>=1.2.1"
      ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


import warnings
import pandas as pd
import numpy as np


class BinaryPerformance:
    def __init__(self, y_true, y_pred):
        """Gives detailed perfomance metrics for binary calssification models.
        
        :param y_true: Array of true targets.
        :type y_true: :obj:`numpy.array`
        :param y_pred: Array of predicted targets.
        :type y_pred: :obj:`numpy.array`
        """
        warnings.filterwarnings('ignore')
        # Put data to a DF
        for_conf = pd.DataFrame({'true': y_true,
            'pred': y_pred,
            'perf': ''})

        # Compute True and False positives/negatives
        for_conf['perf'][(for_conf.true == 1) & (for_conf.pred == 1)] = 'true_positive'
        for_conf['perf'][(for_conf.true == 0) & (for_conf.pred == 0)] = 'true_negative'
        for_conf['perf'][(for_conf.true == 1) & (for_conf.pred == 0)] = 'false_negative'
        for_conf['perf'][(for_conf.true == 0) & (for_conf.pred == 1)] = 'false_positive'
        # Number of True and False positives/negatives
        self.tp = len(for_conf[(for_conf.perf == 'true_positive')])
        self.tn = len(for_conf[(for_conf.perf == 'true_negative')])
        self.fn = len(for_conf[(for_conf.perf == 'false_negative')])
        self.fp = len(for_conf[(for_conf.perf == 'false_positive')])
        self.m = len(y_pred)
        # Confusion matrix
        self.conf = pd.DataFrame({'True 0': [0., 0.], 'True 1': [0., 0.]}, index=['Predicted 0', 'Predicted 1'])
        # Fill confusion matrix
        self.conf['True 0']['Predicted 0'] = self.tn/self.m
        self.conf['True 0']['Predicted 1'] = self.fp/self.m
        self.conf['True 1']['Predicted 0'] = self.fn/self.m
        self.conf['True 1']['Predicted 1'] = self.tp/self.m
        warnings.filterwarnings('default')

    def accuracy(self):
        """Binary accuracy of the model. Percentage of equal values between :obj:`y_true` and :obj:`y_pred`.
        
        :formula: .. math:: accuracy = \dfrac{TP + TN}{n}

        :return: Accuracy
        :rtype: :obj:`float`
        """
        return((self.tp + self.tn)/self.m)

    def confusion(self):
        """Confusion matrix

        :return: Confusion matrix

            +-----------------+------------+------------+
            |                 | **True 0** | **True 0** |
            +=================+============+============+
            | **Predicted 0** | :math:`TN` | :math:`TN` |
            +-----------------+------------+------------+
            | **Predicted 1** | :math:`FP` | :math:`TP` |
            +-----------------+------------+------------+
        
        :rtype: :obj:`pandas.DataFrame`
        """
        return(self.conf * 100)

    def precision(self):
        """Precision metric, proportion of actual 1 values amongst the ones predicted.
        
        :formula: .. math:: precision = \dfrac{TP}{TP + FP}

        :return: Precision
        :rtype: :obj:`float`
        """
        return(self.tp / (self.tp + self.fp))

    def recall(self):
        """  Recall metric, proportion of values we predicted as one from the actuals ones.
        
        :formula: .. math:: recall = \dfrac{TP}{TP + FN}

        :return: Recall
        :rtype: :obj:`float`
        """
        return(self.tp / (self.tp + self.fn))

    def F1_score(self):
        """F1-score
        
        :formula: .. math:: F_{1} = 2 \cdot \dfrac{precision \\times recall}{precision + recall}

        :return: F1-score
        :rtype: :obj:`float`
        """
        return(2 * (self.precision() * self.recall()) / (self.precision() + self.recall()))



import theano
import theano.tensor as T
import numpy as np
import random
import matplotlib.pyplot as plt

import pycof as pc

from .optimizers import Adam, AdaGrad, AdaMax
from .optimizers import SGD, MomentumSGD
from .optimizers import RMSprop
from .initializations import init_params
from .activations import tanh, sigmoid, relu, softplus
from .losses import mean_squared_error, binary_cross_entropy

"""
DeepLearning website - Yoshua Bengio:
- http://deeplearning.net/tutorial/mlp.html
"""

## TODO: add batch
## TODO: comment out

T_activations = {None: None,
                'linear': None,
                'tanh': T.tanh,
                'sigmoid': T.nnet.sigmoid,
                'relu': T.nnet.relu,
                'softplus': T.nnet.softplus,
                }

activations = {None: None,
                'linear': None,
                'tanh': tanh,
                'sigmoid': sigmoid,
                'relu': relu,
                'softplus': softplus,
                }

optimizers = {'sgd': SGD,
                'stochastic_gradient_descent': SGD,
                'momentum': MomentumSGD,
                'momentumsgd': MomentumSGD,
                'adam': Adam,
                'adagrad': AdaGrad,
                'adamax': AdaMax,
                'rmsprop': RMSprop,
                }


class Layer(object):
    def __init__(self, n_in, n_out, W=None, b=None, activation=None, seed=None, init_weights='xavier', init_bias='zeros'):
        """Fully connected Neural Network layer
        
        :param n_in: Dimension of the input layer (number of rows for the weights matrix).
        :type n_in: int
        :param n_out: Dimension of the output layer (number of columns for the weights matrix).
        :type n_out: int
        :param W: Weights matrix, defaults to None (generated by :obj:`init_params` in `initializations <initializations.html#statinf.ml.initializations.init_params>`_).
        :type W: numpy.array, optional
        :param b: Vector of bias, defaults to None (generated by :obj:`init_params` in `initializations <initializations.html#statinf.ml.initializations.init_params>`_).
        :type b: numpy.array, optional
        :param activation: Activation function to be applied, defaults to None (i.e. linear, see `activations <activations.html>`_)
        :type activation: str, optional
        :param seed: Random seed, defaults to None.
        :type seed: int, optional
        :param init_weights: Distribution to be used when initializing the weights matrix, defaults to 'xavier' (see `initializations <initializations.html#statinf.ml.initializations.init_params>`_).
        :type init_weights: str, optional
        :param init_bias: Method to used for initializing the bias vector, defaults to 'zeros' (see `initializations <initializations.html#statinf.ml.initializations.init_params>`_).
        :type init_bias: str, optional

        :return: Layer to be stacked to the Neural Network
        :rtype: Layer
        """
      
        self.input = input
        self.activation = activation
        self.f = T_activations[activation]
        self.activate = activations[activation]

        if W is None:
            # If weights matrix is not provided (default), initialize from a distribution
            W_values = init_params(rows=n_in, cols=n_out, method=init_weights, seed=seed)
            # Multiply by 4 the weights for sigmoid activation. See Y. Bengio's website
            if activation == theano.tensor.nnet.sigmoid:
                W_values *= 4
        elif W.shape == (n_in, n_out):
            raise ValueError(f'Weights dimension does not match. Dimension for W should be {(n_in, n_out)} and got {W.shape}.')

        if b is None:
            # If bias matrix is not provided (default), initialize from a distribution
            b_values = init_params(rows=1, cols=n_out, method=init_bias, seed=seed)
        elif b.shape == (n_in, n_out):
            raise ValueError(f'Weights dimension does not match. Dimension for b should be {(1, n_out)} and got {b.shape}.')
        
        # Store parameters in the class as shared parameters with theano
        self.W = theano.shared(value=W_values, name='W', borrow=True)
        self.b = theano.shared(value=b_values, name='b', borrow=True)
        self.optimal_w = W_values
        self.optimal_b = b_values
        
        # Parameters of the model
        self.params = [self.W, self.b]

    def _feed_forward(self, input, tensor=True):
        if tensor:
            Xb_e = T.dot(input, self.W) + self.b
            self.output = (Xb_e if self.activation in [None, 'linear'] else self.f(Xb_e))
        else:
            # Used as feed forward for prediction (avoid using Theano)
            Xb_e = input.dot(self.W.get_value()) + self.b.get_value()
            self.output = (Xb_e if self.activation in [None, 'linear'] else self.activate(Xb_e))
        return self.output

class MLP:
    """Multi-Layer Perceptron
    
    :references: Rosenblatt, F. (1958). `The perceptron: a probabilistic model for information storage and organization in the brain <http://web.engr.oregonstate.edu/~huanlian/teaching/machine-learning/2017fall/extra/rosenblatt-1958.pdf>`_. Psychological review, 65(6), 386.
    
    """
    def __init__(self, loss='MSE', random=None):
        self.loss = loss
        self._layers = []
        self.params = []
        self.L1 = 0.
        self.L2 = 0.
        self._cost_value = []

    def _L1(self):
        """Compute L1 regularization
        
        :formula: .. math:: L_{1} = \sum_{i=1}^{L} |\\beta|

        :return: L1 penalty
        :rtype: float
        """
        for layer in self._layers:
            # L1 norm ; one regularization option is to enforce L1 norm to
            # be small
            self.L1 += abs(layer.W).sum()
        return self.L1
    
    def _L2(self):
        """Compute L2 regularization
        
        :formula: .. math:: L_{2} = \sum_{i=1}^{L} \\beta^{2}

        :return: L2 penalty
        :rtype: float
        """
        for layer in self._layers:
            # square of L2 norm ; one regularization option is to enforce
            # square of L2 norm to be small
            self.L2 += (layer.W ** 2).sum()
        return self.L2
    
    def _params(self):
        # the parameters of the model are the parameters of the two layer it is
        # made out of
        _params = []
        for layer in self._layers:
            _params += list(layer.params)
        
        self.params = list(_params)
    
    def add(self, layer):
        """Stack layer to Neural Network
        
        :param layer: Layer to be stacked
        :type layer: :func:`~Layer`
        """
        self._layers.append(layer)
    
    def _forward_prop(self, x, tensor=True):
        output = x
        for layer in self._layers:
            output = layer._feed_forward(output, tensor)
        return output

    def _cost(self, x, y):
        output = self._forward_prop(x)
        # Compute loss
        if self.loss.lower() in ['mse', 'mean_squared_error']:
            _cost_value = mean_squared_error(y_true=y, y_pred=output) # ((output - y) ** 2).sum()
        elif self.loss.lower() in ['binary_cross_entropy', 'bce']:
            _cost_value = binary_cross_entropy(y_true=y, y_pred=output)
        else:
            raise ValueError('Loss function is not valid.')
        self._cost_value += [_cost_value]
        return _cost_value
    

    def get_weights(self, layer='all', param='all'):
        """Fetches the parameters from the network.
        
        :param layer: Layer id from which to fetch the parameters, defaults to 'all'
        :type layer: int, optional
        :param param: What parameter we need to fetch (can either be 'weights', 'bias' or 'all'), defaults to 'all'
        :type param: str, optional
        
        :return: Weights and Bias used in the Neural Network.
        :rtype: dict
        """
        # Check what parameter the user was to get
        if param.lower() in ['weights', 'w', 'weight']:
            par = 0
        elif param.lower() in ['biases', 'b', 'bias']:
            par = 1
        elif param.lower() == 'all':
            par = 2
        else:
            raise ValueError('Value for "param" is not value. Please chose between "weights", "bias" or "all".')
        
        if layer == 'all':
            # If user wants to see all layers, we create a dictionnary
            weights = {}
            for i in range(len(self._layers)):
                if par == 0:
                    weights.update({f'weights {i}': self._layers[i].W.get_value()})
                elif par == 1:
                    weights.update({f'bias {i}': self._layers[i].b.get_value()})
                else:
                    weights.update({f'weights {i}': self._layers[i].W.get_value(), f'bias {i}': self._layers[i].b.get_value()})
        elif layer in range(len(self._layers)):
            # If user wants only 1 specific layer,
            if par == 2:
                # we return a dict for all params
                weights = {'weights': self._layers[layer].Z.get_value(), 'bias': self._layers[layer].b.get_value()}
            else:
                # or an array for 1 single param
                weights = self._layers[layer].Z.get_value()
        else:
            raise ValueError(f'Layer is incorrect. Please chose either "all" or layer <= {len(self._layers) - 1}. Got layer = {layer}')
        
        return weights

    def train(self, data, X, Y='Y', epochs=100, optimizer='SGD', batch_size=1, training_size=0.8, test_set=None, learning_rate=0.01, L1_reg=0., L2_reg=0., early_stop=True, patience=10, improvement_threshold=0.995, restore_weights=True, verbose=True, verbose_all=False, plot=False):
        """Train the Neural Network.
        
        :param data: Layer id from which to fetch the parameters.
        :type data: pandas.DataFrame
        :param X: List of X input variables.
        :type X: list
        :param Y: Variable to predict, defaults to 'Y'.
        :type Y: str, optional
        :param epochs: Number of epochs to train the network, defaults to 100.
        :type epochs: int, optional
        :param optimizer: Algortihm to use to minimize the loss function, defaults to 'SGD' (see `optimizers <optimizers.html>`_).
        :type optimizer: str, optional
        :param batch_size: Size of each batch to be trained, defaults to 1.
        :type batch_size: int, optional
        :param training_size: Ratio of the data to be used for training set (:math:`\in (0, 1)`) the remainder is used for test set, defaults to 0.8.
        :type training_size: float, optional
        :param test_set: Data frame to use as test set (overrides **training_size** if provided), defaults to None.
        :type test_set: pandas.DataFrame, optional
        :param learning_rate: Learning rate (step size) for gradient descent, defaults to 0.01.
        :type learning_rate: float, optional
        :param L1_reg: Coefficient :math:`\\lambda_{1} \in (0,1)` for the L1 penalty, defaults to 0.
        :type L1_reg: float, optional
        :param L2_reg: Coefficient :math:`\\lambda_{2} \in (0,1)` for the L2 penalty, defaults to 0.
        :type L2_reg: float, optional
        :param early_stop: Apply early-stopping method to avoid over-fitting. Stop the training sequence after the patience (see below) with no improvement, defaults to True.
        :type early_stop: bool, optional
        :param patience:  Number of iterations to wait after a minimum has been found. If no improvement, training sequence is stopped, defaults to 10.
        :type patience: int, optional
        :param improvement_threshold: Change level of the loss from which we consider it is an improvement, defaults to 0.995.
        :type improvement_threshold: float, optional
        :param restore_weights: If training is early-stopped, restore the weights of the best iteration, defaults to True.
        :type restore_weights: bool, optional
        :param verbose: Verbosity level (True or 1 for minimal display, 2 for details), defaults to True.
        :type verbose: int, optional
        :param plot: Plot the evolution of the training and test loss for each epoch, defaults to False.
        :type plot: bool, optional

        :references: * Bergstra, J., Bastien, F., Breuleux, O., Lamblin, P., Pascanu, R., Delalleau, O., ... & Bengio, Y. (2011). `Theano: Deep learning on gpus with python <http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.678.1889&rep=rep1&type=pdf>`_. In NIPS 2011, BigLearning Workshop, Granada, Spain (Vol. 3, pp. 1-48). Microtome Publishing. 
            * Friedman, J., Hastie, T., & Tibshirani, R. (2001). `The elements of statistical learning <https://web.stanford.edu/~hastie/Papers/ESLII.pdf>`_ (Vol. 1, No. 10). New York: Springer series in statistics.
            * Goodfellow, I., Bengio, Y., & Courville, A. (2016). `Deep learning <https://www.deeplearningbook.org/>`_. MIT press. 
        """
        # Parse arguments to pass to self
        self.learning_rate = learning_rate
        self.L1_reg = L1_reg
        self.L2_reg = L2_reg
        self.X_col = X
        self.Y_col = Y
        self.optimizer = str(optimizer).lower()

        verbose = True if verbose in [True, 1, 2] else False
        verbose_all = True if verbose in [2] else False

        # Get indexes
        self.data_size = data.shape[0]

        if test_set is None:
            data_idx = list(data.index)
            train_index = random.sample(data_idx, int(self.data_size * training_size))
            test_index = [i for i in data_idx if i not in train_index]
            # Train set
            train_set_x = data.loc[train_index, self.X_col]
            train_set_y = data.loc[train_index, self.Y_col]
            # Test set
            test_set_x = data.loc[test_index, self.X_col]
            test_set_y = data.loc[test_index, self.Y_col]
        else:
            # Train set
            train_set_x = data[self.X_col]
            train_set_y = data[self.Y_col]
            # Test set
            test_set_x = test_set[self.X_col]
            test_set_y = test_set[self.Y_col]
        
        # Set data to theano
        # Train set
        train_set_x = theano.shared(np.array(train_set_x, dtype=theano.config.floatX))
        train_set_y = theano.shared(np.array(train_set_y, dtype=theano.config.floatX))
        # Test set
        test_set_x = theano.shared(np.array(test_set_x, dtype=theano.config.floatX))
        test_set_y = theano.shared(np.array(test_set_y, dtype=theano.config.floatX))
        self.test_set = test_set_x

        index = T.lscalar()  # index to a [mini]batch
        x = T.vector('x')
        y = T.scalar('y')

        # Get initial parameters
        self._params()

        # Set cost
        cost = self._cost(x, y) + (self.L1_reg * self._L1()) + (self.L2_reg * self._L2())
        
        ## Define update with optimizer
        try:
            self.updates = optimizers[self.optimizer](params=self.params, learning_rate=self.learning_rate).updates(cost)
        except BaseException:
            posible_opt = "', '".join([i for i in optimizers.keys()])
            raise ValueError(f"Optimizer not value. Please make sure you choose in '{posible_opt}'. Got '{self.optimizer}'.")
 
        # Define Theano function for training step
        train_model = theano.function(
            inputs=[index],
            outputs=cost,
            updates=self.updates,
            givens={x: train_set_x[index], y: train_set_y[index]}
        )
        
        # Define Theano function for validation step
        validate_model = theano.function(
            inputs=[index],
            outputs=cost,
            givens={x: test_set_x[index], y: test_set_y[index]}
        )
        # Define length of test
        n_train_set = train_set_x.get_value(borrow=True).shape[0]
        n_test_set = test_set_x.get_value(borrow=True).shape[0]
        
        # Initialize values
        self.train_losses = []       # Training loss vector to plot
        self.test_losses = []        # Test loss vector to plot
        self.best_loss = np.inf # Initial best loss value
        self.best_iter = 0
        verif_i = epochs        # First, verify epoch at then end (init only)
        i = 0

        # Start training
        while i <= epochs:
            i += 1
            # Measure accuracy on train set
            epoch_loss = np.mean([train_model(i) for i in range(n_train_set)])
            # Evaluate on test set
            epoch_loss_test = np.mean([validate_model(i).tolist() for i in range(n_test_set)])
            # Save to historical performance
            self.train_losses += [epoch_loss]
            self.test_losses += [epoch_loss_test]
            # If we got the best score until now, and patience is not reached
            if (epoch_loss_test < self.best_loss * improvement_threshold) & (i <= verif_i) & early_stop:
                pc.verbose_display(f'New min found for iter {i}', verbose=verbose_all)
                verif_i = i + patience
                self.best_loss = epoch_loss_test
                self.best_iter = i
                # self.optimal_w = self.w.get_value()
                # self.optimal_b = self.b.get_value()
                for layer in self._layers:
                    layer.optimal_w = layer.W.get_value()
                    layer.optimal_b = layer.b.get_value()
            # if no improvement
            elif (i < verif_i):
                # pc.verbose_display(f'No improvement at iter {i}', verbose=verbose_all)
                pass
            else:
                pc.verbose_display(f'Stop training. Minimum found after {self.best_iter} iterations', verbose=verbose)
                if restore_weights:
                    for layer in self._layers:
                        layer.W = theano.shared(value=layer.optimal_w, name='W', borrow=True)
                        layer.b = theano.shared(value=layer.optimal_b, name='b', borrow=True)
                        layer.params = [layer.W, layer.b]
                break

        if plot:
            # Plot the training and test loss
            plt.title('MLP train vs. test error through epochs', loc='center')
            plt.plot(self.train_losses, label='Training loss')
            plt.plot(self.test_losses, label='Test loss')
            plt.legend()
            plt.show()
    

    def predict(self, new_data, binary=False, threshold=0.5):
        """
        """
        # Transform input data to be transformed into numpy array format
        x = new_data[self.X_col].values
        # Apply feedforward step to the data to get the prediction (apply weights and activations)
        output = self._forward_prop(x, tensor=False)
        if binary:
            return [1. if y[0] > threshold else 0. for y in output.reshape((new_data.shape[0], 1))]
        else:
            return [y[0] for y in output.reshape((new_data.shape[0], 1))]
